// Copyright (c) 2009 Steve Gribble.
//
// linked_list.c:   a simple linked list implementation
//
// Note that this linked list implementation is *not* thread-safe!

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "linked_list.h"

// An element inside the linked list.
typedef struct link_data_t {
  list_element_t      e;      // the element that is stored
  struct link_data_t *next;   // a pointer to the next link element
} link_data_t;

// The list itself.
typedef struct list_data_t {
  struct link_data_t *front;  // 1st element in list; NULL if empty
} list_data_t;

linked_list_t linked_list_create() {
  list_data_t *new_list = malloc(sizeof(list_data_t));
  if (new_list == NULL)
    return NULL;

  new_list->front = NULL;
  return (linked_list_t) new_list;
}

void linked_list_destroy(linked_list_t list) {
  list_data_t *linked_list = (list_data_t *) list;
  assert(linked_list != NULL);
  free(linked_list);
}

boolean_t linked_list_add_to_front(linked_list_t list,
                                   list_element_t element) {
  list_data_t *linked_list = (list_data_t *) list;
  link_data_t *new_element;

  // Validate argument.
  if (linked_list == NULL)
    return FALSE;

  // Allocate a new element.
  new_element = (link_data_t *) malloc(sizeof(link_data_t));
  if (new_element == NULL)
    return FALSE;
  new_element->e = element;
  new_element->next = NULL;

  // Is this the first element in the list?
  if (linked_list->front == NULL) {
    linked_list->front = new_element;
    return TRUE;
  }

  // Nope.  Insert into front.
  new_element->next = linked_list->front;
  linked_list->front = new_element;
  return TRUE;
}

boolean_t linked_list_add_to_back(linked_list_t list,
                                  list_element_t element) {
  list_data_t *linked_list = (list_data_t *) list;
  link_data_t *new_element, *cur_element;

  // Validate argument.
  if (linked_list == NULL)
    return FALSE;

  // Allocate a new element.
  new_element = (link_data_t *) malloc(sizeof(link_data_t));
  if (new_element == NULL)
    return FALSE;
  new_element->e = element;
  new_element->next = NULL;

  // Is this the first element in the list?
  if (linked_list->front == NULL) {
    linked_list->front = new_element;
    return TRUE;
  }

  // Nope.  Need to iterate to the end of the list, and insert there.
  cur_element = linked_list->front;
  while (cur_element->next != NULL)
    cur_element = cur_element->next;
  cur_element->next = new_element;
  return TRUE;
}

void linked_list_remove_from_front(linked_list_t list) {
  list_data_t *linked_list = (list_data_t *) list;
  link_data_t *oldfront;

  // Validate argument.
  assert(linked_list != NULL);
  assert(linked_list->front != NULL);

  oldfront = linked_list->front;
  linked_list->front = (linked_list->front)->next;
  free(oldfront);
}

void linked_list_remove_from_back(linked_list_t list) {
  list_data_t *linked_list = (list_data_t *) list;
  link_data_t *element, *prev_element;

  // Validate argument.
  assert(linked_list != NULL);
  assert(linked_list->front != NULL);

  // If there is only one element in the list, remove it.
  if ((linked_list->front)->next == NULL) {
    free(linked_list->front);
    linked_list->front = NULL;
    return;
  }

  // Iterate to the back of the list, keeping track of previous element
  // as well.
  prev_element = linked_list->front;
  element = prev_element->next;
  while (element->next != NULL) {
    prev_element = element;
    element = element->next;
  }

  // Remove the last element.
  free(prev_element->next);
  prev_element->next = NULL;
}

list_element_t linked_list_front(linked_list_t list) {
  list_data_t *linkedlist = (list_data_t *) list;
  assert(linkedlist != NULL);
  assert(linkedlist->front != NULL);

  return (linkedlist->front)->e;
}

list_element_t linked_list_back(linked_list_t list) {
  list_data_t *linkedlist = (list_data_t *) list;
  link_data_t *element;

  assert(linkedlist != NULL);
  assert(linkedlist->front != NULL);

  // iterate to the end of the list.
  element = linkedlist->front;
  while (element->next != NULL)
    element = element->next;

  return element->e;

}

boolean_t is_list_empty(linked_list_t list) {
  list_data_t *linkedlist = (list_data_t *) list;

  assert(linkedlist != NULL);
  if (linkedlist->front == NULL)
    return TRUE;
  return FALSE;
}
